/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <modalcv.h>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>

// input files
#define TEST_IMAGE_1		"/usr/share/libmodalcv/ov9782/left.jpg"
#define TEST_IMAGE_2		"/usr/share/libmodalcv/ov9782/right.jpg"

// output files
#define OUTPUT_FOLDER	"/data/libmodalcv/"
#define OUTPUT_FILE		"/data/libmodalcv/dcm.png"

#define LOOPS 100


static int64_t _apps_time_monotonic_ns()
{
    struct timespec ts;
    if(clock_gettime(CLOCK_MONOTONIC, &ts)){
        fprintf(stderr,"ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}


int main()
{
    mcv_en_debug_prints(1);

	if(system("rm -f /data/libmodalcv/dcm/*"))
	{
		fprintf(stderr, "Warning, failed to delete old files from %s\n", OUTPUT_FOLDER);
		return -1;
	}
    if(mkdir(OUTPUT_FOLDER, S_IRWXU) && errno!=EEXIST){
        fprintf(stderr, "failed to make folder\n");
        return -1;
    }

    // load in images
    printf("loading images from disk\n");
    cv::Mat img_1, img_2;
    img_1 = imread(TEST_IMAGE_1, cv::IMREAD_GRAYSCALE);
    img_2 = imread(TEST_IMAGE_2, cv::IMREAD_GRAYSCALE);

    int w = img_1.cols;
    int h = img_1.rows;

    if (img_1.type() != CV_8UC1 || img_2.type() != CV_8UC1) {
        fprintf(stderr, "ERROR loading images\n");
        return -1;
    }

    // configure and run feature extraction to have points for NCC
    mcv_cvp_fpx_config_t fpx_config;
    fpx_config.width           = w;
    fpx_config.height          = h;
    fpx_config.mode            = MCV_CVP_FPX_PEAK_8x8;
    fpx_config.nms_mode        = MCV_CVP_FPX_5_TAP_NMS;
    fpx_config.score_threshold = 3000;
    fpx_config.robustness      = 50;                     // 0-127, default 10

    // create fpx instances to get points for DCM
    mcv_cvp_fpx_handle fpx_handle_1 = mcv_cvp_fpx_init(fpx_config);
    mcv_cvp_fpx_handle fpx_handle_2 = mcv_cvp_fpx_init(fpx_config);
    if(fpx_handle_1 == NULL || fpx_handle_2 == NULL)
    {
        fprintf(stderr, "ERROR getting FPX handle\n");
        return -1;
    }

    // create output objects
    mcv_cvp_fpx_t fpx_1, fpx_2;
    if(mcv_cvp_create_fpx(fpx_handle_1, &fpx_1))
    {
        fprintf(stderr, "ERROR creating FPX output object\n");
    }
    if(mcv_cvp_create_fpx(fpx_handle_2, &fpx_2))
    {
        fprintf(stderr, "ERROR creating FPX output object\n");
    }
    
    // extract features
    printf("extracting features\n");
    for(int i = 0; i < 2; i++)
    {
        if(mcv_cvp_fpx_process(fpx_handle_1, img_1.data, &fpx_1))
        {
            fprintf(stderr, "ERROR completing FPX process\n");
            return -1;
        }
        if(mcv_cvp_fpx_process(fpx_handle_2, img_2.data, &fpx_2))
        {
            fprintf(stderr, "ERROR completing FPX process\n");
            return -1;
        }
    }

    // configure DCM and get handle
    mcv_cvp_dcm_config_t dcm_config;
    dcm_config.w = w;
    dcm_config.h = h;

    mcv_cvp_dcm_handle dcm_handle = mcv_cvp_dcm_init(dcm_config);
    if(NULL == dcm_handle)
    {
        fprintf(stderr, "ERROR getting DCM handle\n");
        return -1;
    } 

    // create an output object
    mcv_cvp_dcm_output_t dcm_output;
    if(mcv_cvp_create_dcm_output(dcm_handle, &dcm_output))
    {
        fprintf(stderr, "ERROR creating DCM output object\n");
        return -1;
    }

    // for first image, just compute descriptors
    if(mcv_cvp_dcm_process(dcm_handle, img_1.data, &fpx_1, NULL, 0, &dcm_output))
    {
        fprintf(stderr, "ERROR computing DCM for image 1\n");
        return -1;
    }


    // now compute desc for second image and match with output desc from
    // first image
    int64_t start = _apps_time_monotonic_ns();
    if(mcv_cvp_dcm_process(dcm_handle, img_2.data, &fpx_2, dcm_output.descriptors, 
        dcm_output.n_descriptors, &dcm_output))
    {
        fprintf(stderr, "ERROR computing DCM for image 2\n");
        return -1;
    }
    int64_t end = _apps_time_monotonic_ns();
    printf("DCM Time ::    %6.2fms\n", (double)(end-start)/1000000);

    // draw descriptor matches between the two
    cv::Mat final_img;
    cv::hconcat(img_1, img_2, final_img);
    cv::cvtColor(final_img, final_img, cv::COLOR_GRAY2BGR);
    for(int match_idx = 0; match_idx < dcm_output.n_matches; match_idx++)
    {
        mcv_cvp_fpx_feature_t feature_2 = fpx_2.features[match_idx];
        uint16_t best_feature_match_idx = dcm_output.matches[match_idx].index;
        uint16_t score = dcm_output.matches[match_idx].score;
        // printf("best matched feature %d -> %d (%d)\n", match_idx, best_feature_match_idx, score);
        mcv_cvp_fpx_feature_t feature_1 = fpx_1.features[best_feature_match_idx];

        if(score < 15)
        {
            cv::line(final_img, cv::Point2f(feature_1.x, feature_1.y), cv::Point2f(feature_2.x + w, feature_2.y),
                cv::Scalar(0,255,0), 1, 1);
        }
    }
    cv::imwrite(OUTPUT_FILE, final_img);

    // cleanup
    mcv_cvp_destroy_fpx(fpx_handle_1, &fpx_1);
    mcv_cvp_destroy_fpx(fpx_handle_2, &fpx_2);
    mcv_cvp_destroy_dcm_output(dcm_handle, &dcm_output);

    return 0;
}