/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/**
 * @file       mcv-test-fpx.cpp <examples/mcv-test-undistort>
 *
 * @brief      Tests the Feature Extraction (FPX) process
 * 
 * @author     Thomas Patton (thomas.patton@modalai.com)
 * @date       2023
 */

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <modalcv.h>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>

using namespace cv;

// input files (640x480)
// #define TEST_IMAGE		"/usr/share/libmodalcv/dfs/left.png"

// input files (1280x800)
#define TEST_IMAGE		"/usr/share/libmodalcv/J2119/00900.png"

// output files
#define OUTPUT_FOLDER	"/data/libmodalcv/"
#define OUTPUT_FILE		"/data/libmodalcv/fpx.png"

#define LOOPS 100


static int64_t _apps_time_monotonic_ns()
{
    struct timespec ts;
    if(clock_gettime(CLOCK_MONOTONIC, &ts)){
        fprintf(stderr,"ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}


int main()
{
    int ret, i;

    // make output directory to save debug images to
    if(mkdir(OUTPUT_FOLDER, S_IRWXU) && errno!=EEXIST){
        fprintf(stderr, "failed to make folder\n");
        return -1;
    }

    // load in image
    printf("loading image from disk\n");
    Mat img_mat = imread(TEST_IMAGE, IMREAD_GRAYSCALE);
    if(img_mat.type()!=CV_8UC1){
        fprintf(stderr, "ERROR loading %s\n", TEST_IMAGE);
        return -1;
    }
    int width = img_mat.size().width;
    int height = img_mat.size().height;
    printf("width: %3d height: %3d\n", width, height);

    // initialize feature point extractor
    mcv_cvp_fpx_config_t fpx_config;
    fpx_config.width           = width;
    fpx_config.height          = height;
    fpx_config.mode            = MCV_CVP_FPX_PEAK_8x8;
    fpx_config.nms_mode        = MCV_CVP_FPX_5_TAP_NMS;
    fpx_config.score_threshold = 80;
    fpx_config.robustness      = 50;                     // 0-127, default 10

    // get a handle for our FPX process
    mcv_cvp_fpx_handle fpx_handle = mcv_cvp_fpx_init(fpx_config);
    if(fpx_handle == NULL)
    {
        fprintf(stderr, "ERROR getting FPX handle\n");
        return -1;
    }

    // create output object
    mcv_cvp_fpx_t fpx;
    if(mcv_cvp_create_fpx(fpx_handle, &fpx))
    {
        fprintf(stderr, "ERROR creating FPX output object\n");
    }

    // extract features
    printf("extracting features\n");
    int64_t t0 = _apps_time_monotonic_ns();
    for(int i=0; i<LOOPS; i++){
        if(mcv_cvp_fpx_process(fpx_handle, img_mat.data, &fpx))
        {
            fprintf(stderr, "ERROR completing FPX process\n");
            return -1;
        }
    }
    int64_t t1 = _apps_time_monotonic_ns();

    printf("n_points:  %5d\n", fpx.n_points);
    printf("max_score: %5d\n", fpx.max_score);

    // print timing
    int64_t t_per_loop_ns = (t1-t0)/LOOPS;
    printf("%6.2fms   feature extraction time\n", ((double)(t_per_loop_ns))/1000000.0);

    // draw overlay
    for(i = 0; i < (int) fpx.n_points; i++){
        int x = fpx.features[i].x;
        int y = fpx.features[i].y;
        mcv_overlay_draw_plus(img_mat.data, width, height, x, y);
        // printf("xy: %3d %3d score: %10d\n", x, y, fpx_output.features[i].score);
    }

    // save to disk
    printf("saving image to disk\n");
    imwrite(OUTPUT_FILE, img_mat);

    // cleanup
    mcv_cvp_fpx_deinit(fpx_handle);

    printf("setting permissions for output files\n");
    ret = system("chmod 777 -R /data/libmodalcv/");
    if(ret){
        fprintf(stderr, "Warning, failed to set permissions for output files\n");
    }

    printf("DONE\n");
    return 0;
}
