/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <modalcv.h>

#include <opencv2/imgcodecs.hpp>


// input files
#define TEST_IMAGE		"/usr/share/libmodalcv/dfs/left.png"

// output files
#define OUTPUT_FOLDER	"/data/libmodalcv/pyr/"
#define LOOPS	100 // run test 100 times for timing purposes


static int64_t _apps_time_monotonic_ns()
{
	struct timespec ts;
	if(clock_gettime(CLOCK_MONOTONIC, &ts)){
		fprintf(stderr,"ERROR calling clock_gettime\n");
		return -1;
	}
	return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}


int main()
{
	mcv_en_debug_prints(0);
	mcv_en_timing_prints(1);

	// make output directory to save debug images to
	if(mkdir(OUTPUT_FOLDER, S_IRWXU) && errno!=EEXIST){
		fprintf(stderr, "failed to make folder\n");
		return -1;
	}

	// load in image
	printf("loading image from disk\n");
	cv::Mat img_mat;
	img_mat = cv::imread(TEST_IMAGE, cv::IMREAD_GRAYSCALE);
	if(img_mat.type()!=CV_8UC1){
		fprintf(stderr, "ERROR loading %s\n", TEST_IMAGE);
		return -1;
	}

	// initialize pyramid extractor
	mcv_cvp_pyr_config_t pyr_config;
	pyr_config.width               = img_mat.cols;
	pyr_config.height              = img_mat.rows;
	pyr_config.n_octaves           = 3;
	pyr_config.n_scales_per_octave = 1;
    memset(pyr_config.n_filter_coeffs, 3, MCV_CVP_MAX_PYRAMID_OCTAVES*sizeof(uint32_t));

	mcv_cvp_pyr_handle handle = mcv_cvp_pyr_init(pyr_config);
	if(NULL == handle)
	{
		fprintf(stderr, "ERROR creating pyramid handle\n");
		return -1;
	}

	// create output structure
	mcv_cvp_pyr_t pyr;
	mcv_cvp_create_pyr(handle, &pyr);

	// perform pyramid extraction
	printf("extracting image pyramid\n");
	int64_t t0 = _apps_time_monotonic_ns();
	for(int i = 0; i < LOOPS ;i++)
	{
		if(mcv_cvp_pyr_process(handle, img_mat.data, &pyr))
		{
			fprintf(stderr, "ERROR executing pyramid process\n");
			return -1;
		}
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   avg pyramid extraction time\n", ((double)(t_per_loop_ns))/1000000.0);

	// save to disk
	printf("saving images to disk\n");
    for (int i = 0; i < (int) pyr_config.n_octaves; i++){
        cv::Mat curr_pyr(pyr.height[i], pyr.width[i], CV_8UC1, 
			pyr.image_pyramids[i]);
        std::stringstream ss;
        ss << OUTPUT_FOLDER << "pyramid_" << i << ".png";
	    imwrite(ss.str(), curr_pyr);
    }

	// ////////////////////////////////////////////////////////////////////////////
	// cleanup
	////////////////////////////////////////////////////////////////////////////
	if(mcv_cvp_destroy_pyr(handle, &pyr))
	{
		fprintf(stderr, "ERROR destorying pyramid object\n");
		return -1;
	}
	if(mcv_cvp_pyr_deinit(handle))
	{
		fprintf(stderr, "ERROR deinitializing pyramid\n");
		return -1;
	}

	printf("DONE\n");
	return 0;
}
